/**********************************************************************/
/* The 1st IPSJ SIG-ARC High-Performance Processor Design Contest     */
/* 310_sort Version 2013-10-13                                        */
/**********************************************************************/

#include "../../lib/lib.c"
/**********************************************************************/
/* Input Data */
volatile int *init_edges= (int*)0x00040000;
volatile int *init_n    = (int*)0x0007fff0;
volatile int *init_m    = (int*)0x0007fff4;
volatile int *init_start= (int*)0x0007fff8;
volatile int *init_goal = (int*)0x0007fffc;

/**********************************************************************/
/* Definitions and Working data */

// Max number node is 2048. "1" added because the node number is 1-origin.
#define MAX_NODE 2048+1

// Max integer is the max positive 32-bit number : 0x7FFF_FFFF
#define MAX_INT 2147483647

// True and false logical values
#define TRUE  1
#define FALSE 0

// message flags for debug and verbose purposes
#define debug FALSE
#define verbose FALSE

typedef struct {
  int currentCost;
  int isMinimumCost;
  int fromNodeForMinimumCost;
} node;

typedef struct {
  int from;
  int to;
  int cost;
} edge;

int numberOfEdges;
int numberOfNodes;
edge* edgeArray;
node nodeArray [MAX_NODE];

/**********************************************************************/
/* Print Functions */

/**
 * Print edge array information for debug
 */
void printEdgeArray()
{
  int i;

  mylib_display_char('#');
  mylib_display_char('E');
  mylib_display_char('d');
  mylib_display_char('g');
  mylib_display_char('e');
  mylib_display_char('s');
  mylib_display_newline();

  for(i=0; i<numberOfEdges; i++) {
    edge* e = &edgeArray[i];

    mylib_display_char('[');
    mylib_display_dec(i);
    mylib_display_char(']');
    mylib_display_char('\t');
    mylib_display_dec(e->from);
    mylib_display_char('\t');
    mylib_display_dec(e->to);
    mylib_display_char('\t');
    mylib_display_dec(e->cost);
    mylib_display_newline();
  }
}

/**
 * Print node array for debug
 */
void printNodeArray()
{
  mylib_display_char('#');
  mylib_display_char('N');
  mylib_display_char('o');
  mylib_display_char('d');
  mylib_display_char('e');
  mylib_display_char('s');
  mylib_display_newline();

  int i;
  for(i=1; i<=numberOfNodes; i++) {
    node* n = &nodeArray[i];

    mylib_display_char('[');
    mylib_display_dec(i);
    mylib_display_char(']');
    mylib_display_char('\t');
    mylib_display_dec(n->currentCost);
    mylib_display_char('\t');
    mylib_display_dec(n->isMinimumCost);
    mylib_display_char('\t');
    mylib_display_dec(n->fromNodeForMinimumCost);
    mylib_display_newline();
  }
}

/**
 * Print initial message 
 */
void printInitialMessage(int n, int m, int start, int goal) {
  mylib_display_char('s');
  mylib_display_char('p');
  mylib_display_char('a');
  mylib_display_char('t');
  mylib_display_char('h');
  mylib_display_char(' ');
  mylib_display_char('n');
  mylib_display_char('=');
  mylib_display_dec(n);
  mylib_display_char(' ');
  mylib_display_char('m');
  mylib_display_char('=');
  mylib_display_dec(m);
  mylib_display_char(' ');
  mylib_display_char('s');
  mylib_display_char('t');
  mylib_display_char('a');
  mylib_display_char('r');
  mylib_display_char('t');
  mylib_display_char('=');
  mylib_display_dec(start);
  mylib_display_char(' ');
  mylib_display_char('g');
  mylib_display_char('o');
  mylib_display_char('a');
  mylib_display_char('l');
  mylib_display_char('=');
  mylib_display_dec(goal);
  mylib_display_newline();
}

/**
 * Print result message
 * @param sumOfCost sum of the cost along the route
 * @param start start node of the route
 * @param goal goal node of the route
 */
void printResultMessage(int sumOfCost, int start, int goal) {
  int currentNode = start;

  mylib_display_newline();
  mylib_display_char('r');
  mylib_display_char('o');
  mylib_display_char('u');
  mylib_display_char('t');
  mylib_display_char('e');
  mylib_display_char('=');
  while(TRUE) {
    mylib_display_dec(currentNode);
    if(currentNode == 0) break;
    if(currentNode == goal) break;

    currentNode = nodeArray[currentNode].fromNodeForMinimumCost;
    mylib_display_char(',');
  }
  mylib_display_newline();

  mylib_display_char('c');
  mylib_display_char('o');
  mylib_display_char('s');
  mylib_display_char('t');
  mylib_display_char('=');
  mylib_display_dec(sumOfCost);
  mylib_display_newline();
}

/**********************************************************************/
/* Dijkstra Algorithm Functions */
/**
 * Find current shortest node
 *
 * @return current shortest node#
 */
int findCurrentShortestNode() {
  int i;
  int currentShortestCost = MAX_INT;
  int currentShortestNode = 0;

  // loop for find the minimum cost for all nodes
  for(i=1; i<=numberOfNodes; i++) {
    node* n = &nodeArray[i];
    int isMinimumCost = n->isMinimumCost;
    if(!isMinimumCost) {
      int currentCost = n->currentCost;
      if(currentShortestCost > currentCost) {
        currentShortestCost = currentCost;
        currentShortestNode = i;
      }
    }
  }

  return currentShortestNode;
}


/**
 * Update next shortest nodes
 *
 */
void updateNextShortestNodes()
{
  int i;
  // update all edges
  for(i=0; i<numberOfEdges; i++) {
    edge* e = &edgeArray[i];
    int from = e->from;
    int to   = e->to;
    int cost = e->cost;
    node* fromNode = &nodeArray[from];
    node* toNode = &nodeArray[to];
    int fromNodeIsMinimumCost = fromNode->isMinimumCost;
    int toNodeIsMinimumCost   = toNode->isMinimumCost;

    // update "toNode" which is next of "fromNode"
    if(fromNodeIsMinimumCost) {
      if(!toNodeIsMinimumCost) {
        int currentCostOfTheNextNode = toNode->currentCost;
        int updatedCostOfTheNextNode = fromNode->currentCost + cost;
        // compares the current and the updated value
        if(updatedCostOfTheNextNode < currentCostOfTheNextNode) {
          nodeArray[to].currentCost = updatedCostOfTheNextNode;
          nodeArray[to].fromNodeForMinimumCost = from;
        }
      }
    }
    // update "fromNode" which is the next of "toNode"
    if(toNodeIsMinimumCost) {
      if(!fromNodeIsMinimumCost) {
        int currentCostOfTheNextNode = fromNode->currentCost;
        int updatedCostOfTheNextNode = toNode->currentCost + cost;
        // compares the current and the updated value
        if(updatedCostOfTheNextNode < currentCostOfTheNextNode) {
          nodeArray[from].currentCost = updatedCostOfTheNextNode;
          nodeArray[from].fromNodeForMinimumCost = to;
        }
      }
    }
  }
}

/**
 * Find next shortest node
 *
 * @return newly found shortest node#
 */
int findNextShortestNode() {
  int nextShortestNode;

  updateNextShortestNodes();
  nextShortestNode = findCurrentShortestNode();
  node* n = &nodeArray[nextShortestNode];
  n->isMinimumCost = TRUE;

  return nextShortestNode;
}


/**
 * Find shortest path
 * result is output to serial
 *
 * @param start start node#
 * @param goal  goal node#
 * @return Sum of the cost along the shortest path
 */
int findShortestPath(int start, int goal)
{
  int newShortestNode = -1;
  int currentShortestNode = -1;
  if(start < 0 || goal < 0) {
    return MAX_INT;
  }

  nodeArray[goal].currentCost = 0;
  nodeArray[goal].isMinimumCost = TRUE;
  if(debug) printNodeArray(); // initial state

  while(newShortestNode != start) {
    newShortestNode = findNextShortestNode();
    if(debug) printNodeArray(); // updated state
    if(verbose) mylib_display_char('.');

    if(currentShortestNode == newShortestNode) break;
    currentShortestNode = newShortestNode;
  }

  if(verbose) mylib_display_newline();
  return nodeArray[start].currentCost;
}


/**********************************************************************/
/* main and initialize functions */

/**
 * Initialize node array information
 * @param n number of nodes
 */
void initNodeArray(int n) {
  int i;
  for(i=1; i<=n; i++) {
    nodeArray[i].currentCost   = MAX_INT;
    nodeArray[i].isMinimumCost = FALSE;
  }
}


/**
 * Shortest PATH main function
 *
 * @return status code
 */
int main(void)
{
  int sumOfCost = 0;
  int n = *init_n;
  int m = *init_m;
  int start = *init_start;
  int goal  = *init_goal;
  printInitialMessage(n, m, start, goal);

  /* ----- Initialize ----- */
  edgeArray = (edge*) init_edges;
  numberOfNodes = n;
  numberOfEdges = m;

  if(debug) printEdgeArray();
  initNodeArray(n);
  if(debug) printNodeArray();

  /* -----  main kernel ----- */
  sumOfCost = findShortestPath(start, goal);

  /* -----  show result ----- */
  printResultMessage(sumOfCost, start, goal);

  mylib_display_newline();

  mylib_finalize();
  return 0;
}
/**********************************************************************/
